import json

from flask import request


class IndexAPI:
    @staticmethod
    def get():
        url = request.url
        return json.dumps({
            "links": [
                url + 'company',
                url + 'client',
                url + 'staff',
                url + 'status',
                url + 'project',
                url + 'job',
                url + 'invoice'
            ]
        })



class Action:
    def __init__(self, key, label, action=None):
        self.key = key
        self.label = label
        self.action = action

    def __repr__(self):
        return self.label

    def check_input(self, input_value):
        return input_value == self.key

    def execute(self):
        if self.action:
            self.action()



from invoicing.api.base_api import BaseAPI
from invoicing.models.client_model import ClientModel
from invoicing.repository.client_repository import ClientRepository


class ClientAPI(BaseAPI):

    def __init__(self):
        super().__init__(ClientRepository(), ClientModel())



import json


class JSONTransformer:
    @staticmethod
    def resultsToJSON(rows, headers):
        results = []
        for row in rows:
            data = {}
            for i, header in enumerate(headers):
                data[header] = row[i]
            results.append(data)
        return JSONTransformer.dataToJSON(results)

    @staticmethod
    def resultToJSON(row, headers):
        result = {}
        for i, header in enumerate(headers):
            result[header] = row[i]
        return JSONTransformer.dataToJSON(result)

    @staticmethod
    def dataToJSON(data):
        response = {
            "data": data
        }
        return json.dumps(response)

    @staticmethod
    def errorToJSON(errors):
        response = {
            "errors": errors
        }
        return json.dumps(response)

    @staticmethod
    def messageToJSON(message):
        response = {
            "message": message
        }
        return json.dumps(response)



from invoicing.crud.base_crud import BaseCrud
from invoicing.latex.latex_invoice import LatexInvoice
from invoicing.models.invoice_model import InvoiceModel
from invoicing.repository.invoice_repository import InvoiceRepository
from invoicing.repository.job_repository import JobRepository
from invoicing.ui.date import Date
from invoicing.ui.menu import Menu
from invoicing.ui.style import Style
from invoicing.value_validation.value_validation import Validation


class InvoiceCrud(BaseCrud):
    def __init__(self):
        super().__init__('Invoices', InvoiceRepository, InvoiceModel)
        self.menu_actions.add_action('Generate', self.generate)

    def make_paginated_menu(self):
        return self.paginated_menu(
            find=self.repository.find_paginated_join_clients_and_companies,
            find_by_id=self.repository.find_by_id_join_clients_and_companies
        )

    def generate(self):
        print(Style.create_title('Generate Invoice'))
        invoice = self.make_paginated_menu()
        if invoice:
            jobRepository = JobRepository()
            jobs = jobRepository.find_jobs_by_invoice_id(invoice['id'])
            self.enter_billable_time(jobRepository, jobs)
            jobs = jobRepository.find_jobs_by_invoice_id(invoice['id'])
            invoice_data = self.make_invoice_dictionary(invoice, jobs)
            LatexInvoice().generate(**invoice_data)
            self.mark_invoiced_jobs_as_complete(jobRepository, jobs)
            Menu.wait_for_input()

    def enter_billable_time(self, jobRepository, jobs):
        print(Style.create_title('Enter Billable Time'))
        for job in jobs:
            print('Title: %s' % job['title'])
            print('Description: %s' % job['description'])
            print('Estimated Time: %s' % job['estimated_time'])
            print('Logged Time: %s' % job['actual_time'])
            billable = ''
            while not Validation.is_float(billable):
                billable = input('Billable Time: ')
            jobRepository.update_billable_time(job['id'], billable)
            jobRepository.save()
            jobRepository.check_rows_updated('Job Updated')

    def make_invoice_dictionary(self, invoice, jobs):
        invoice_data = {
            'reference_code': invoice['reference_code'],
            'company_name': invoice['company_name'],
            'company_address': invoice['company_address'],
            'created_at': Date().convert_date_time_for_printing(invoice['created_at']),
            'total_cost': str(sum([float(job['rate']) * float(job['billable_time']) for job in jobs])),
            'jobs': [{
                'title': job['title'],
                'description': job['description'],
                'type': 'hours',
                'billable_time': str(job['billable_time']),
                'staff_rate': str(job['rate']),
                'cost': str(float(job['rate']) * float(job['billable_time']))
            } for job in jobs]
        }
        return invoice_data

    def mark_invoiced_jobs_as_complete(self, jobRepository, jobs):
        if len(jobs):
            for job in jobs:
                jobRepository.update_mark_as_complete(job['id'])
            jobRepository.save()
            jobRepository.check_rows_updated('The selected jobs have been marked as completed')




from invoicing.crud.base_crud import BaseCrud
from invoicing.models.status_model import StatusModel
from invoicing.repository.status_repository import StatusRepository


class StatusCrud(BaseCrud):
    def __init__(self):
        super().__init__('Statuses', StatusRepository, StatusModel)



from invoicing.ui.menu import Menu
from invoicing.ui.pagination import Pagination
from invoicing.ui.style import Style


class Relationship:

    @staticmethod
    def select_foreign_key_relationship(relationship):
        repository = relationship.repository()
        print(Style.create_title('Select %s' % relationship.name))
        paginated_menu = relationship.paginated_menu or Pagination(repository)
        item = paginated_menu()
        if item:
            return item[0]
        else:
            return False

    @staticmethod
    def select_foreign_key_relationship_inverse(relationship):
        foreign_keys = []
        repository = relationship.repository()
        while Menu.yes_no_question('Add %s' % relationship.name):
            print(Style.create_title('Select %s' % relationship.name))
            paginated_menu = relationship.paginated_menu or Pagination(repository)
            item = paginated_menu()
            if item:
                foreign_keys.append(item['id'])
            else:
                print('\nNo relationship added\n')
        return relationship.related_name, foreign_keys



from abc import ABCMeta

from ansi_colours import AnsiColours as Colour

from invoicing.actions.action_collection import ActionCollection
from invoicing.model_validation.field import ForeignKeyField, OneToManyField
from invoicing.ui.menu import Menu
from invoicing.ui.pagination import Pagination
from invoicing.ui.style import Style
from invoicing.ui.relationship import Relationship


class BaseCrud(metaclass=ABCMeta):
    __metaclass__ = ABCMeta

    def __init__(self, table_name, repository, model):
        self.table_name = table_name
        self.repository = repository()
        self.model = model()
        self.paginated_menu = Pagination(self.repository)
        self.menu_actions = ActionCollection(
            ('View', self.show),
            ('Add', self.add),
            ('Edit', self.edit),
            ('Delete', self.delete)
        )

    def menu(self):
        Menu.create('Manage ' + self.table_name, self.menu_actions)

    def show(self):
        print(Style.create_title('Show %s' % self.table_name))
        item = self.make_paginated_menu()
        if item:
            self.show_item_detail(item)
            self.show_item_menu(item['id'])

    def show_item_detail(self, item):
        print(Style.create_title('%s Data' % self.table_name))
        for header in self.repository.get_headers():
            label = Style.make_label(header)
            print('%s: %s' % (label, Colour.blue(str(item[header]))))

    def show_item_menu(self, id):
        Menu.create(self.table_name + ' Menu', ActionCollection())

    def add(self):
        print(Style.create_title('Add %s' % self.table_name))
        data = self.input_add_data()
        self.validate_model_data(data['own'])
        if self.model.is_valid():
            self.save_data(data)
        else:
            self.display_errors('%s not added' % self.table_name)
        Menu.wait_for_input()

    def input_add_data(self):
        data = {'own': {}, 'relation': {}}
        for (key, field) in self.model:
            if field.initial_value is not None:
                data['own'][key] = field.initial_value
            elif isinstance(field, ForeignKeyField):
                foreign_key = Relationship.select_foreign_key_relationship(field.relationship)
                if foreign_key:
                    data['own'][key] = foreign_key
                else:
                    break
            elif isinstance(field, OneToManyField):
                foreign_key = Relationship.select_foreign_key_relationship_inverse(field.relationship)
                if foreign_key:
                    data['relation'][key] = foreign_key
                else:
                    break
            elif field.default_value is not None:
                user_input = input("%s (%s): " % (Style.make_label(key), field.default_value))
                if user_input is '':
                    user_input = field.default_value
                data['own'][key] = user_input
            else:
                data['own'][key] = input("%s: " % Style.make_label(key))
        return data

    def save_data(self, data):
        self.repository.insert(data['own'])
        self.repository.update_parent_foreign_keys(data['relation'])
        self.repository.save()
        self.repository.check_rows_updated('%s Added' % self.table_name)

    def edit(self):
        print(Style.create_title('Edit %s' % self.table_name))
        item = self.make_paginated_menu()
        if item:
            data = self.input_edit_data(item)
            data['own'] = self.validate_model_data(data['own'])
            if self.model.is_valid:
                self.save_update(data, item)
            else:
                self.display_errors('%s not updated' % self.table_name)
        else:
            print('No changes made')
        Menu.wait_for_input()

    def input_edit_data(self, item):
        data = {'own': {}, 'relation': {}}
        for (key, field) in self.model:
            if not field.updatable:
                continue
            if isinstance(field, ForeignKeyField):
                if Menu.yes_no_question('Change %s?' % field.relationship.name):
                    foreign_key = Relationship.select_foreign_key_relationship(field.relationship)
                    if foreign_key:
                        data['own'][key] = foreign_key
                else:
                    data['own'][key] = item[key]
            elif isinstance(field, OneToManyField):
                foreign_key = Relationship.select_foreign_key_relationship_inverse(field.relationship)
                if foreign_key:
                    data['relation'][key] = foreign_key
                else:
                    data['relation'][key] = item[key]
                # Todo: handle removing relationships
            else:
                data['own'][key] = self.update_field(item[key], Style.make_label(key))
        return data

    def validate_model_data(self, data):
        self.model(**data)
        self.model.validate()
        return data

    def save_update(self, data, item):
        self.repository.update(item['id'], data['own'])
        self.repository.update_parent_foreign_keys(data['relation'])
        self.repository.save()
        self.repository.check_rows_updated('%s Updated' % self.table_name)

    def update_field(self, current_value, field_name):
        value = input(field_name + "(" + str(current_value) + "): ")
        new_value = value if len(value) > 0 else current_value
        return new_value

    def display_errors(self, message):
        print(Style.create_title(message))
        for (key, value) in self.model.get_errors().items():
            print("%s: %s" % (key.capitalize(), Colour.red(value)))

    def delete(self):
        print(Style.create_title('Delete %s' % self.table_name))
        item = self.make_paginated_menu()
        if item:
            user_action = False
            while not user_action == 'delete':
                user_action = input('Type \'%s\' to remove this item or %s to cancel: ' % (
                    Colour.red('delete'),
                    Colour.green('c')
                ))
                if user_action == 'c':
                    return
            self.save_remove(item)
            Menu.wait_for_input()

    def save_remove(self, item):
        self.repository.remove(item['id'])
        self.repository.save()
        self.repository.check_rows_updated('%s Deleted' % self.table_name)

    def make_paginated_menu(self):
        return self.paginated_menu()




from invoicing.crud.base_crud import BaseCrud
from invoicing.models.company_model import CompanyModel
from invoicing.repository.company_repository import CompanyRepository


class CompanyCrud(BaseCrud):
    def __init__(self):
        super().__init__('Companies', CompanyRepository, CompanyModel)



from ansi_colours import AnsiColours as Colour

from invoicing.actions.action_collection import ActionCollection
from invoicing.ui.style import Style
from invoicing.ui.table import Table
from invoicing.value_validation.value_validation import Validation


class Menu:
    @staticmethod
    def create(title, action_collection: ActionCollection):
        user_input = 0
        keys = []
        while True:
            print(Style.create_title(title))
            for action in action_collection.actions:
                print(Colour.green(action.key) + ": " + str(action))
                keys.append(action.key)
            while user_input not in keys:
                user_input = input('\nSelect an option: ')
            for action in action_collection.actions:
                if action.check_input(user_input):
                    action.execute()
            if user_input == 'b' or user_input == 'q':
                break
            else:
                user_input = 0

    @staticmethod
    def choose_item_by_id(find_by_id):
        item = False
        while not item:
            id = input('\nEnter an id to view or \'b\' to go back: ')
            if id == 'b':
                return False
            if not Validation.is_number(id):
                print('Not a valid number')
                continue
            item = find_by_id(id)
        return item

    @staticmethod
    def select_row(rows, headers, find_by_id):
        Table.create_table(rows, headers)
        return Menu.choose_item_by_id(find_by_id)

    @staticmethod
    def yes_no_question(question):
        response = False
        while response not in ['y', 'Y', '', 'n', 'N']:
            response = input(question + ' (Y/n): ')
        return response in ['Y', 'y', '']

    @staticmethod
    def wait_for_input(text=None):
        input('\nPress %s to continue.' % Colour.green('enter') if not text else "\n%s" % text)



from datetime import datetime

from dateutil.relativedelta import relativedelta

from invoicing.model_validation.field import StringField, FloatField, Field, ForeignKeyField, \
    DateField
from invoicing.model_validation.validations import IsRequired, IsString
from invoicing.models.base_model import BaseModel
from invoicing.models.project_model import ProjectModel
from invoicing.models.staff_model import StaffModel
from invoicing.models.status_model import StatusModel
from invoicing.relationships.base_relationship import BaseRelationship
from invoicing.repository.project_repository import ProjectRepository
from invoicing.repository.staff_repository import StaffRepository
from invoicing.repository.status_repository import StatusRepository


class JobModel(BaseModel):
    project_id = ForeignKeyField(
        BaseRelationship('Project', ProjectRepository, ProjectModel)
    )
    title = StringField([IsRequired()])
    description = StringField([IsRequired()])
    estimated_time = FloatField([IsRequired()])
    deadline = DateField([IsRequired()],
                         default_value=(datetime.today() + relativedelta(months=1)).strftime("%d-%m-%y"))
    assigned_to = ForeignKeyField(
        BaseRelationship('Staff', StaffRepository, StaffModel)
    )
    status_id = ForeignKeyField(
        BaseRelationship('Status', StatusRepository, StatusModel)
    )
    created_at = Field([IsString()], initial_value=datetime.now().strftime("%Y-%m-%d %H:%M"), updatable=False)



from invoicing.query_builder.query_builder import QueryBuilder
from invoicing.repository.base_repository import BaseRepository


class JobRepository(BaseRepository):

    def __init__(self):
        super().__init__('job')

    def find_all_join_staff_and_status(self):
        query = QueryBuilder(self.table) \
            .select(['job.id', 'reference_code', 'job.title', 'status.title as status', 'deadline',
                     '(staff.first_name || \' \' || staff.last_name) as assigned_to',
                     'created_at', 'completed']) \
            .from_() \
            .join('staff', 'assigned_to = staff.id') \
            .join('status', 'status_id = status.id')
        self.execute(**query.build())
        return self.get_all()

    def find_paginated_join_staff_and_status(self, limit=5, page=1):
        query = QueryBuilder(self.table) \
            .select(['job.id', 'reference_code', 'job.title', 'status.title as status', 'deadline',
                     '(staff.first_name || \' \' || staff.last_name) as assigned_to',
                     'created_at', 'completed']) \
            .from_() \
            .join('staff', 'assigned_to = staff.id') \
            .join('status', 'status_id = status.id') \
            .limit(limit) \
            .offset(page * limit - limit)
        self.execute(**query.build())
        return self.get_all()

    def find_by_assigned_to(self, staff_id):
        query = QueryBuilder(self.table) \
            .select(['job.id', 'reference_code', 'job.title', 'description', 'estimated_time', 'actual_time',
                     'status.title as status', 'deadline']) \
            .from_() \
            .join('status', 'status_id = status.id') \
            .where('assigned_to = ?', staff_id)
        self.execute(**query.build())
        return self.get_all()

    def find_paginated_by_assigned_to(self, staff_id, limit=5, page=1):
        query = QueryBuilder(self.table) \
            .select(['job.id', 'reference_code', 'job.title', 'description', 'estimated_time', 'actual_time',
                     'status.title as status', 'deadline']) \
            .from_() \
            .join('status', 'status_id = status.id') \
            .where('assigned_to = ?', staff_id) \
            .limit(limit) \
            .offset(page * limit - limit)
        self.execute(**query.build())
        return self.get_all()

    def find_jobs_by_project_id(self, project_id):
        query = QueryBuilder(self.table) \
            .select(['job.id', 'reference_code', 'job.title as title', 'description', 'estimated_time',
                     'status.title as status', 'deadline',
                     '(staff.first_name || \' \' || staff.last_name) as staff_name', 'staff.rate as rate']) \
            .from_() \
            .join('staff', 'assigned_to = staff.id') \
            .join('status', 'status_id = status.id') \
            .where('job.project_id = ?', project_id)
        self.execute(**query.build())
        return self.get_all()

    def find_jobs_by_invoice_id(self, invoice_id):
        query = QueryBuilder(self.table) \
            .select(
            ['job.id', 'reference_code', 'job.title as title', 'description', 'billable_time', 'estimated_time',
             'actual_time',
                     'status.title as status', 'deadline',
                     '(staff.first_name || \' \' || staff.last_name) as staff_name', 'staff.rate as rate']) \
            .from_() \
            .join('staff', 'assigned_to = staff.id') \
            .join('status', 'status_id = status.id') \
            .where('job.invoice_id = ?', invoice_id)
        self.execute(**query.build())
        return self.get_all()

    def find_jobs_by_client_id_where_complete(self, client_id):
        query = QueryBuilder(self.table) \
            .select(['job.id', 'job.reference_code', 'job.title',
                     'status.title as status', 'deadline',
                     '(staff.first_name || \' \' || staff.last_name) as staff_name']) \
            .from_() \
            .join('staff', 'assigned_to = staff.id') \
            .join('status', 'status_id = status.id') \
            .join('project', 'project_id = project.id') \
            .where('project.client_id = ?', client_id) \
            .and_where('completed = ?', '1')
        self.execute(**query.build())
        return self.get_all()

    def find_paginated_jobs_by_client_id_where_not_complete(self, client_id, limit=5, page=1):
        query = QueryBuilder(self.table) \
            .select(['job.id', 'job.reference_code', 'job.title',
                     'status.title as status', 'deadline',
                     '(staff.first_name || \' \' || staff.last_name) as staff_name']) \
            .from_() \
            .join('staff', 'assigned_to = staff.id') \
            .join('status', 'status_id = status.id') \
            .join('project', 'project_id = project.id') \
            .where('project.client_id = ?', client_id) \
            .and_where('completed = ?', '0') \
            .limit(limit) \
            .offset(limit * page - limit)
        self.execute(**query.build())
        return self.get_all()

    def find_last_reference_code(self):
        query = QueryBuilder(self.table) \
            .select(['reference_code as last_reference_code']) \
            .from_() \
            .where('id = (select max(id) from job)')
        self.execute(**query.build())
        return self.get_one()

    def make_next_reference_code(self):
        last_job = self.find_last_reference_code()
        last_reference_code = last_job['last_reference_code'] if last_job else 'J-1000'
        reference_code = 'J-' + str(int(last_reference_code[2:]) + 1)
        return reference_code

    def update_actual_time(self, id, time_spent):
        query = QueryBuilder(self.table) \
            .raw('update job set actual_time = (actual_time + ?) where id = ?', (time_spent, id))
        self.execute(**query.build())

    def update_billable_time(self, id, time_spent):
        query = QueryBuilder(self.table) \
            .update({'billable_time': time_spent}) \
            .where('id = ?', id)
        self.execute(**query.build())

    def update_mark_as_complete(self, id):
        query = QueryBuilder(self.table) \
            .update({'completed': 1}) \
            .where('id = ?', id)
        self.execute(**query.build())

    def add_to_invoice(self, id, invoice_id):
        query = QueryBuilder(self.table) \
            .update({'invoice_id': invoice_id}) \
            .where('id = ?', id)
        self.execute(**query.build())

    def remove_jobs_by_project_id(self, project_id):
        query = QueryBuilder(self.table) \
            .delete() \
            .where('project_id = ?', project_id)
        self.execute(**query.build())

    def remove_jobs_by_invoice_id(self, invoice_id):
        query = QueryBuilder(self.table) \
            .delete() \
            .where('invoice_id = ?', invoice_id)
        self.execute(**query.build())



from invoicing.model_validation.field import StringField
from invoicing.model_validation.validations import IsRequired
from invoicing.models.base_model import BaseModel


class CompanyModel(BaseModel):
    # Todo: find way all fields to be null or empty if not required
    name = StringField([IsRequired()])
    address = StringField()



class QueryBuilder:
    def __init__(self, table):
        self.table = table
        self.query = ''
        self.params = []

    def select(self, params):
        self.query += 'select {} '.format(', '.join(params))
        return self

    def from_(self, table=None):
        table = table if table else self.table
        self.query += 'from {} '.format(table)
        return self

    def where(self, condition, param=None):
        self.add_param(param)
        self.query += 'where {} '.format(condition)
        return self

    def limit(self, limit):
        self.add_param(limit)
        self.query += 'limit ? '
        return self

    def offset(self, offset):
        self.add_param(offset)
        self.query += 'offset ? '
        return self

    def count(self):
        self.query += 'select count(*) '
        return self

    def where_id_is_last_inserted(self):
        self.query += 'where id = last_insert_rowid() '
        return self

    def and_where(self, condition, param=None):
        self.add_param(param)
        self.query += 'and {} '.format(condition)
        return self

    def join(self, table, on):
        self.query += 'join {} on {} '.format(table, on)
        return self

    def insert(self, values):
        for k, v in values.items():
            self.add_param(v)
        self.query += 'insert into {} ({}) values ({}) '.format(
            self.table,
            ', '.join([k for k in values]),
            ', '.join(['?' for _ in range(len(values))])
        )
        return self

    def update(self, set):
        for k, v in set.items():
            self.add_param(v)
        self.query = 'update {} set {} '.format(self.table, ', '.join([k + ' = ?' for k in set]))
        return self

    def delete(self, table=None):
        table = table if table else self.table
        self.query += 'delete from {} '.format(table)
        return self

    def add_param(self, param):
        if param != None:
            self.params.append(param)

    def raw(self, query, params=None):
        for param in params:
            self.add_param(param)
        self.query += query + ' '
        return self

    def build(self):
        return {'sql': self.query, 'parameters': tuple(self.params)}



from invoicing.model_validation.field import FloatField, StringField
from invoicing.model_validation.validations import IsRequired
from invoicing.models.base_model import BaseModel


class StaffModel(BaseModel):
    first_name = StringField([IsRequired()])
    last_name = StringField([IsRequired()])
    job_title = StringField([IsRequired()])
    rate = FloatField([IsRequired()])



import json
from datetime import datetime

from invoicing.api.base_api import BaseAPI
from invoicing.models.job_model import JobModel
from invoicing.repository.job_repository import JobRepository
from invoicing.transformer.json import JSONTransformer


class JobAPI(BaseAPI):

    def __init__(self):
        super().__init__(JobRepository(), JobModel())

    def add(self, data):
        self.model(**data)
        self.model.validate()
        if self.model.is_valid:
            data["reference_code"] = self.repository.make_next_reference_code()
            data["created_at"] = datetime.now().strftime("%Y-%m-%d %H:%M")
            self.repository.insert(data)
            self.repository.save()
            return JSONTransformer.resultToJSON(self.repository.find_last_inserted(), self.repository.get_headers())
        else:
            return json.dumps(self.model.get_errors())



from invoicing.query_builder.query_builder import QueryBuilder
from invoicing.repository.base_repository import BaseRepository


class InvoiceRepository(BaseRepository):

    def __init__(self):
        super().__init__('invoice')

    def find_all_join_clients_and_companies(self):
        query = QueryBuilder(self.table) \
            .select(['invoice.id', 'reference_code', 'created_at',
                     'client.fullname as client_fullname',
                     'company.name as company_name']) \
            .from_() \
            .join('client', 'client_id = client.id') \
            .join('company', 'client.company_id = company.id')
        self.execute(**query.build())
        return self.get_all()

    def find_paginated_join_clients_and_companies(self, limit=5, page=1):
        query = QueryBuilder(self.table) \
            .select(['invoice.id', 'reference_code', 'created_at',
                     'client.id as client_id, client.fullname as client_fullname',
                     'company.name as company_name']) \
            .from_() \
            .join('client', 'client_id = client.id') \
            .join('company', 'client.company_id = company.id') \
            .limit(limit) \
            .offset(page * limit - limit)
        self.execute(**query.build())
        return self.get_all()

    def find_by_id_join_clients_and_companies(self, id):
        query = QueryBuilder(self.table) \
            .select(['invoice.id as id', 'reference_code', 'created_at',
                     'client.id as client_id, client.fullname as client_fullname',
                     'company.name as company_name', 'company.address as company_address']) \
            .from_() \
            .join('client', 'client_id = client.id') \
            .join('company', 'client.company_id = company.id') \
            .where('invoice.id = ?', id)
        self.execute(**query.build())
        return self.get_one()

    def find_invoices_by_client_id(self, client_id):
        query = QueryBuilder(self.table) \
            .select(['id', 'reference_code', 'created_at']) \
            .from_() \
            .where('client_id = ?', client_id)
        self.execute(**query.build())
        return self.get_all()

    def find_last_reference_code(self):
        query = QueryBuilder(self.table) \
            .select(['id', 'reference_code as last_reference_code']) \
            .from_() \
            .where('id = (select max(id) from invoices)')
        self.execute(**query.build())
        return self.get_one()

    def remove_invoices_by_client_id(self, client_id):
        query = QueryBuilder(self.table) \
            .delete() \
            .where('client_id = ?', client_id)
        self.execute(**query.build())



class Validation:
    @staticmethod
    def is_number(number):
        try:
            int(number)
        except ValueError:
            return False
        return True

    @staticmethod
    def is_float(number):
        try:
            float(number)
        except ValueError:
            return False
        return True

    @staticmethod
    def is_number_and_in_range(number, start, stop):
        if Validation.is_number(number) and number >= start and number <= stop:
            return True
        return False



from invoicing.model_validation.validations import IsInteger, IsFloat, IsString, IsBoolean, IsDate, IsCallable
from invoicing.model_validation.validator import Validator
from invoicing.relationships.base_relationship import BaseRelationship, OneToManyRelationship


class Field(Validator):
    def __init__(self, validation_links=None, **kwargs):
        super().__init__()
        kwargs.setdefault('initial_value', None)
        kwargs.setdefault('default_value', None)
        kwargs.setdefault('updatable', True)
        validation_links = validation_links if validation_links is not None else []
        if validation_links:
            self.set_validation_links(validation_links)
        self.initial_value = kwargs.get('initial_value')
        self.default_value = kwargs.get('default_value')
        self.value = kwargs.get('initial_value')
        self.updatable = kwargs.get('updatable')

    def set_value(self, value):
        self.value = value

    def __str__(self):
        return "Value: %s: %s (%s)" % (self.value, self.error_message, 'valid' if self.is_valid() else 'invalid')


class IntegerField(Field):
    def __init__(self, validation_links=None, **kwargs):
        super().__init__(validation_links=validation_links, **kwargs)
        self.set_validation_link(IsInteger(), front=True)

    def set_value(self, value):
        try:
            self.value = int(value)
        except:
            self.value = value


class DateField(Field):
    def __init__(self, validation_links=None, **kwargs):
        super().__init__(validation_links=validation_links, **kwargs)
        self.set_validation_link(IsDate(), front=True)
        self.set_validation_link(IsString(), front=True)

    def set_value(self, value):
        try:
            self.value = str(value)
        except:
            self.value = value


class StringField(Field):
    def __init__(self, validation_links=None, **kwargs):
        super().__init__(validation_links=validation_links, **kwargs)
        self.set_validation_link(IsString(), front=True)

    def set_value(self, value):
        try:
            self.value = str(value)
        except:
            self.value = value


class BooleanField(Field):
    def __init__(self, validation_links=None, **kwargs):
        super().__init__(validation_links=validation_links, **kwargs)
        self.set_validation_link(IsBoolean(), front=True)

    def set_value(self, value):
        try:
            self.value = bool(value)
        except:
            self.value = value


class FloatField(Field):
    def __init__(self, validation_links=None, **kwargs):
        super().__init__(validation_links=validation_links, **kwargs)
        self.set_validation_link(IsFloat(), front=True)

    def set_value(self, value):
        try:
            self.value = float(value)
        except:
            self.value = value


class ForeignKeyField(Field):
    def __init__(self, relationship: BaseRelationship, validation_links=None, **kwargs):
        super().__init__(validation_links=validation_links, **kwargs)
        self.relationship = relationship
        self.set_validation_link(IsInteger(), front=True)

    def set_value(self, value):
        try:
            self.value = int(value)
        except:
            self.value = value


class OneToManyField(Field):
    def __init__(self, relationship: OneToManyRelationship, validation_links=None, **kwargs):
        super().__init__(validation_links=validation_links, **kwargs)
        self.relationship = relationship
        self.set_validation_link(IsCallable(), front=True)

    def set_value(self, value):
        self.value = value



from ansi_colours import AnsiColours as Colour


class Table:
    @staticmethod
    def create_table(rows, headers):
        col_widths = [len(header) for header in headers]
        for row in rows:
            for i, column in enumerate(row):
                if len(str(column)) > col_widths[i]:
                    col_widths[i] = len(str(column))

        print(Table.make_row(headers, col_widths))
        print(Table.make_underline(col_widths))
        for row in rows:
            print(Table.make_row(row, col_widths, has_key=True))

    @staticmethod
    def make_row(row, col_widths, separator="|", has_key=False):
        string = ""
        for i, column in enumerate(row):
            data = str(column)
            if i is 0 and has_key:
                data = Colour.green(data)
            padding = "".join([" " for _ in range(col_widths[i] - len(str(column)) + 2)])
            string += "  " + data + padding + Colour.light_grey(separator)
        return string

    @staticmethod
    def make_underline(col_widths):
        underline = ""
        for width in col_widths:
            underline += "".join(["-" for _ in range(width + 5)])
        return Colour.light_grey(underline)



from invoicing.models.base_model import BaseModel
from invoicing.repository.base_repository import BaseRepository
from invoicing.transformer.json import JSONTransformer


# Todo: Paginate results
# Todo: handle saving relationships
class BaseAPI:

    def __init__(self, repository: BaseRepository, model: BaseModel):
        self.model = model
        self.repository = repository

    def show_all(self):
        return JSONTransformer.resultsToJSON(self.repository.find_all(), self.repository.get_headers())

    def show_by_id(self, id):
        return JSONTransformer.resultToJSON(self.repository.find_by_id(id), self.repository.get_headers())

    def update_by_id(self, id, data):
        self.model(**data)
        self.model.validate()
        if self.model.is_valid():
            self.repository.update(id, data)
            self.repository.save()
            return JSONTransformer.resultToJSON(self.repository.find_by_id(id), self.repository.get_headers())
        else:
            return JSONTransformer.errorToJSON(self.model.get_errors())

    def add(self, data):
        self.model(**data)
        self.model.validate()
        if self.model.is_valid:
            self.repository.insert(data)
            self.repository.save()
            return JSONTransformer.resultToJSON(self.repository.find_last_inserted(), self.repository.get_headers())
        else:
            return JSONTransformer.errorToJSON(self.model.get_errors())

    def delete(self, id):
        # Todo: make sure id exists and deletion has actually occurred
        self.repository.remove(id)
        self.repository.save()
        return JSONTransformer.messageToJSON("deleted: %s" % id)



from datetime import datetime


class Date:
    @staticmethod
    def convert_date_for_saving(date):
        if not len(date) > 0 or date == 'DD-MM-YYYY':
            return ''
        date = datetime.strptime(date, "%d-%m-%Y")
        return date.strftime('%Y-%m-%d')

    @staticmethod
    def convert_date_for_printing(date):
        if not len(date) > 0:
            return ''
        date = datetime.strptime(date, '%Y-%m-%d')
        return date.strftime('%d-%m-%Y')

    @staticmethod
    def convert_date_time_for_printing(date):
        if not len(date) > 0:
            return ''
        date = datetime.strptime(date, '%Y-%m-%d %H:%M')
        return date.strftime('%d/%m/%Y %H:%M')



"""
This is an alternative entry point for a Flask HTTP API it uses the api modules.
"""

from flask import Flask, request

from invoicing.api.client_api import ClientAPI
from invoicing.api.company_api import CompanyAPI
from invoicing.api.index_api import IndexAPI
from invoicing.api.invoice_api import InvoiceAPI
from invoicing.api.job_api import JobAPI
from invoicing.api.project_api import ProjectAPI
from invoicing.api.staff_api import StaffAPI
from invoicing.api.status_api import StatusAPI

app = Flask(__name__)


def json_response(json):
    response = app.response_class(
        response=json,
        status=200,
        mimetype='application/json'
    )
    return response


class RequestHandler:
    @staticmethod
    def list(api):
        if request.method == 'POST':
            return json_response(api.add(request.get_json()))
        else:
            return json_response(api.show_all())

    @staticmethod
    def detail(api, id):
        if request.method == 'PUT':
            return json_response(api.update_by_id(id, request.get_json()))
        elif request.method == 'DELETE':
            return json_response(api.delete(id))
        else:
            return json_response(api.show_by_id(id))


@app.route('/')
def index():
    return json_response(IndexAPI.get())


@app.route('/company', methods=['GET', 'POST'])
def company_list():
    return RequestHandler.list(CompanyAPI())


@app.route('/company/<id>', methods=['GET', 'PUT', 'DELETE'])
def company_detail(id):
    return RequestHandler.detail(CompanyAPI(), id)


@app.route('/client', methods=['GET', 'POST'])
def client_list():
    return RequestHandler.list(ClientAPI())


@app.route('/client/<id>', methods=['GET', 'PUT', 'DELETE'])
def client_detail(id):
    return RequestHandler.detail(ClientAPI(), id)


@app.route('/staff', methods=['GET', 'POST'])
def staff_list():
    return RequestHandler.list(StaffAPI())


@app.route('/staff/<id>', methods=['GET', 'PUT', 'DELETE'])
def staff_detail(id):
    return RequestHandler.detail(StaffAPI(), id)


@app.route('/status', methods=['GET', 'POST'])
def status_list():
    return RequestHandler.list(StatusAPI())


@app.route('/status/<id>', methods=['GET', 'PUT', 'DELETE'])
def status_detail(id):
    return RequestHandler.detail(StatusAPI(), id)


@app.route('/project', methods=['GET', 'POST'])
def project_list():
    return RequestHandler.list(ProjectAPI())


@app.route('/project/<id>', methods=['GET', 'PUT', 'DELETE'])
def project_detail(id):
    return RequestHandler.detail(ProjectAPI(), id)


@app.route('/job', methods=['GET', 'POST'])
def job_list():
    return RequestHandler.list(JobAPI())


@app.route('/job/<id>', methods=['GET', 'PUT', 'DELETE'])
def job_detail(id):
    return RequestHandler.detail(JobAPI(), id)


@app.route('/invoice', methods=['GET', 'POST'])
def invoice_list():
    return RequestHandler.list(InvoiceAPI())


@app.route('/invoice/<id>', methods=['GET', 'PUT', 'DELETE'])
def invoice_detail(id):
    return RequestHandler.detail(InvoiceAPI(), id)



from abc import ABCMeta

from invoicing.database.sqlite3 import Sqlite3Database
from invoicing.query_builder.query_builder import QueryBuilder


class BaseRepository(Sqlite3Database, metaclass=ABCMeta):
    __metaclass__ = ABCMeta

    def __init__(self, table):
        super().__init__()
        self.table = table

    def find_all(self, select='*'):
        query = QueryBuilder(self.table) \
            .select(select) \
            .from_()
        self.execute(**query.build())
        return self.get_all()

    def find_paginated(self, select='*', limit=10, page=1):
        query = QueryBuilder(self.table) \
            .select(select) \
            .from_() \
            .limit(limit) \
            .offset(page * limit - limit)
        self.execute(**query.build())
        return self.get_all()

    def find_by_id(self, id, select='*'):
        query = QueryBuilder(self.table) \
            .select(select) \
            .from_() \
            .where('id = ?', id)
        self.execute(**query.build())
        return self.get_one()

    def get_count(self):
        query = QueryBuilder(self.table) \
            .count() \
            .from_()
        self.execute(**query.build())
        return self.get_one()

    def find_last_inserted(self, select='*'):
        query = QueryBuilder(self.table) \
            .select(select) \
            .from_() \
            .where_id_is_last_inserted()
        self.execute(**query.build())
        return self.get_one()

    def insert(self, values):
        query = QueryBuilder(self.table) \
            .insert(values)
        self.execute(**query.build())

    def update_parent_foreign_keys(self, related_keys):
        """
        :param related_keys:  {"related_table": ("related_name", [1, 2, 3])}
        :return: []
        """
        last_id = self.cursor.lastrowid
        for related_table, relations in related_keys.items():
            for id in relations[1]:
                query = QueryBuilder(related_table) \
                    .update({relations[0]: last_id}) \
                    .where('id = ?', id)
                self.execute(**query.build())

    def update(self, id, set):
        query = QueryBuilder(self.table) \
            .update(set) \
            .where('id = ?', id)
        self.execute(**query.build())

    def remove(self, id):
        query = QueryBuilder(self.table) \
            .delete() \
            .where('id = ?', id)
        self.execute(**query.build())



class ValidationLink():
    def __init__(self, chain, validation, front=False):
        self.validation = validation
        self.chain = chain
        if not front:
            self.chain.append(self)
        else:
            self.chain.insert(0, self)

    def next(self):
        location = self.chain.index(self)
        if not self.end():
            return self.chain[location + 1]

    def end(self):
        return (self.chain.index(self) + 1 >= len(self.chain))

    def __call__(self, field):
        validator = self.validation(field)
        if not validator.is_valid() or self.end(): return validator
        return self.next()(field)



import os
import sqlite3
from abc import ABCMeta

from invoicing.settings import DB


class Sqlite3Database(metaclass=ABCMeta):
    __metaclass__ = ABCMeta

    def __init__(self):
        if not os.path.isfile(DB):
            raise OSError("Template was not found: " + DB)
        self.connection = sqlite3.connect(DB)
        self.connection.row_factory = sqlite3.Row
        self.cursor = self.connection.cursor()

    def __exit__(self):
        self.connection.close()

    def has_updated_rows(self):
        return self.connection.total_changes > 0

    def save(self):
        self.connection.commit()

    def get_all(self):
        self.connection.commit()
        return self.cursor.fetchall()

    def get_one(self):
        self.connection.commit()
        return self.cursor.fetchone()

    def get_headers(self):
        return list(map(lambda x: x[0], self.cursor.description))

    def check_rows_updated(self, message):
        if self.has_updated_rows() > 0:
            print('\n' + message)
        else:
            print('\nError performing update, please try again')

    def get_description(self):
        return self.cursor.description

    def execute(self, sql='', parameters=()):
        self.cursor.execute(sql, parameters)



from ansi_colours import AnsiColours as Colour


class Style:
    @staticmethod
    def create_title(title):
        return "\n" + title + "\n" + Style.create_underline(title)

    @staticmethod
    def create_underline(title):
        return Colour.blue("".join(['-' for _ in range(len(title))]))

    @staticmethod
    def make_label(header):
        return " ".join([word.capitalize() for word in header.split('_')])



from invoicing.api.base_api import BaseAPI
from invoicing.models.company_model import CompanyModel
from invoicing.repository.company_repository import CompanyRepository


class CompanyAPI(BaseAPI):

    def __init__(self):
        super().__init__(CompanyRepository(), CompanyModel())



from invoicing.api.base_api import BaseAPI
from invoicing.models.invoice_model import InvoiceModel
from invoicing.repository.invoice_repository import InvoiceRepository


class InvoiceAPI(BaseAPI):

    def __init__(self):
        super().__init__(InvoiceRepository(), InvoiceModel())



from invoicing.model_validation.field import StringField
from invoicing.model_validation.validations import IsRequired
from invoicing.models.base_model import BaseModel


class StatusModel(BaseModel):
    title = StringField([IsRequired()])
    colour = StringField([IsRequired()])



from invoicing.model_validation.validation_link import ValidationLink


class Validator():
    nullable = False
    value = None

    def __init__(self):
        self.valid = False
        self.validation_links = []
        self.error_message = None

    def is_valid(self):
        return self.valid

    def set_valid(self, is_valid):
        self.valid = is_valid

    def is_nullable(self):
        return self.nullable

    def is_null(self):
        return self.value is None

    def set_validation_links(self, validation_links):
        for validation_link in validation_links:
            ValidationLink(self.validation_links, validation_link)

    def set_validation_link(self, validation_link, front=False):
        ValidationLink(self.validation_links, validation_link, front)

    def set_error_message(self, message):
        self.error_message = message

    def get_error_message(self):
        return self.error_message

    def validate(self):
        self.error_message = None
        return self.validation_links[0](self)



from invoicing.actions.action_collection import ActionCollection
from invoicing.crud.base_crud import BaseCrud
from invoicing.models.job_model import JobModel
from invoicing.repository.job_repository import JobRepository
from invoicing.repository.status_repository import StatusRepository
from invoicing.ui.menu import Menu
from invoicing.ui.pagination import Pagination
from invoicing.ui.style import Style
from invoicing.value_validation.value_validation import Validation


class JobCrud(BaseCrud):
    def __init__(self):
        super().__init__('Jobs', JobRepository, JobModel)

    def show_item_menu(self, id):
        Menu.create(self.table_name + ' Menu', ActionCollection(
            ('Update Status', lambda: self.update_status(id)),
            ('Log Time', lambda: self.log_time(id))
        ))

    def make_paginated_menu(self):
        return self.paginated_menu(
            find=self.repository.find_paginated_join_staff_and_status,
            find_by_id=lambda id: self.repository.find_by_id(id, (
                'id', 'reference_code', 'title', 'description',
                'estimated_time', 'actual_time', 'deadline',
                'assigned_to', 'status_id', 'project_id'
            ))
        )

    def log_time(self, job_id):
        logged_time = ''
        while not Validation.is_float(logged_time):
            logged_time = input('Log Time: ')
        self.repository.update_actual_time(job_id, logged_time)
        self.repository.save()
        self.repository.check_rows_updated('Job Updated')
        Menu.wait_for_input()

    def update_status(self, job_id):
        statusRepository = StatusRepository()
        paginated_menu = Pagination(statusRepository)
        status = paginated_menu(
            find=statusRepository.find_paginated,
            find_by_id=statusRepository.find_by_id
        )
        self.repository.update(job_id, {
            'status_id': status['id'],
        })
        self.repository.save()
        self.repository.check_rows_updated('Status Updated')
        Menu.wait_for_input()

    def edit_billable_time(self, job):
        print("Estimated Time: " + job['estimated_time'])
        print("Actual Time: " + job['actual_time'])
        billable_time = input("Billable Time: ")
        self.repository.update_billable_time(job['id'], billable_time)
        self.repository.save()
        self.repository.check_rows_updated('Job Updated')
        Menu.wait_for_input()

    def show_jobs_by_assigned_to(self, staff_id):
        print(Style.create_title('Select job to log time'))
        job = self.paginated_menu(
            find=lambda limit, page: self.repository.find_paginated_by_assigned_to(staff_id, limit, page),
            find_by_id=self.repository.find_by_id
        )
        if job:
            self.show_item_detail(job)
            self.show_item_menu(job['id'])



from datetime import datetime

from invoicing.model_validation.field import ForeignKeyField, Field, StringField
from invoicing.model_validation.validations import IsString
from invoicing.models.base_model import BaseModel
from invoicing.models.company_model import CompanyModel
from invoicing.relationships.base_relationship import BaseRelationship
from invoicing.repository.company_repository import CompanyRepository


class ProjectModel(BaseModel):
    title = StringField()
    created_at = Field([IsString()], initial_value=datetime.now().strftime("%Y-%m-%d %H:%M"), updatable=False)
    company_id = ForeignKeyField(
        BaseRelationship('Company', CompanyRepository, CompanyModel)
    )
    # Todo: add one to many relationship type



class BaseRelationship:
    def __init__(self, name, repository, model, paginated_menu=None):
        self.name = name
        self.repository = repository
        self.model = model
        self.paginated_menu = paginated_menu


class OneToManyRelationship(BaseRelationship):
    def __init__(self, related_name, name, repository, model, paginated_menu=None):
        super().__init__(name, repository, model, paginated_menu)
        self.related_name = related_name



from invoicing.actions.action_collection import ActionCollection
from invoicing.crud.base_crud import BaseCrud
from invoicing.crud.job_crud import JobCrud
from invoicing.models.staff_model import StaffModel
from invoicing.repository.staff_repository import StaffRepository
from invoicing.ui.menu import Menu


# Todo: show jobs assigned to staff member
# Todo: Log time against staff members jobs
class StaffCrud(BaseCrud):
    def __init__(self):
        super().__init__('Staff', StaffRepository, StaffModel)

    def show_item_menu(self, id):
        Menu.create(self.table_name + 'Menu', ActionCollection(
            ('Show Assigned Jobs', lambda: JobCrud().show_jobs_by_assigned_to(id)),
        ))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from invoicing.actions.action import Action
from invoicing.actions.action_collection import ActionCollection
from invoicing.controller.main_controller import MainController
from invoicing.ui.menu import Menu


class Invoicing:
    def __init__(self):
        controller = MainController()
        Menu.create('Manage', ActionCollection(
            ('Companies', controller.company_action),
            ('Clients', controller.client_action),
            ('Staff', controller.staff_action),
            ('Statuses', controller.status_action),
            ('Projects', controller.project_action),
            ('Jobs', controller.job_action),
            ('Invoices', controller.invoice_action),
            exit_action=Action('q', 'Quit', False)
        ))


if __name__ == '__main__':
    Invoicing()



import os
import re
import subprocess
from pathlib import Path

import jinja2


class LatexTemplating:

    def __init__(self):
        self.latex_jinja_env = jinja2.Environment(
            block_start_string='\BLOCK{',
            block_end_string='}',
            variable_start_string='\VAR{',
            variable_end_string='}',
            comment_start_string='\#{',
            comment_end_string='}',
            line_statement_prefix='%%',
            line_comment_prefix='%#',
            trim_blocks=True,
            autoescape=False,
            loader=jinja2.FileSystemLoader(os.path.abspath('.'))
        )

    def create_pdf(self, file_name):
        cmd = 'xelatex -interaction=nonstopmode -halt-on-error "' + file_name + '.tex"'
        return_code = subprocess.call(
            cmd,
            shell=True,
            stdout=open('stdout.txt', 'wb'),
            stderr=open('stderr.txt', 'wb')
        )
        try:
            if not return_code == 0:
                self.remove_generated_pdf_if_exists(file_name)
                raise ValueError('\nError {} executing command: {}'.format(return_code, cmd))
            else:
                self.remove_generation_files(file_name)
                print("\n" + file_name + ".pdf has been generated\n")
        except ValueError as e:
            print("\nError creating pdf: " + str(e))
            print("This is probably an issue with the LaTeX file.\n"
                  "Make sure you have LaTeX installed and try to "
                  "compile the template manually.\n")

    def remove_generation_files(self, file_name):
        # os.unlink(file_name + '.tex')
        os.unlink(file_name + '.aux')
        os.unlink(file_name + '.log')

    def remove_generated_pdf_if_exists(self, file_name):
        output = Path(file_name + '.pdf')
        if output.is_file():
            os.unlink(file_name + '.pdf')

    def create_tex_file(self, tex, file_name):
        file = open(file_name + ".tex", "w")
        file.write(tex)
        file.close()

    def tex_escape(self, text):
        """
            :param text: a plain text message
            :return: the message escaped to appear correctly in LaTeX
        """
        conversions = {
            '&': r'\&',
            '%': r'\%',
            '$': r'\$',
            '#': r'\#',
            '_': r'\_',
            '{': r'\{',
            '}': r'\}',
            '~': r'\textasciitilde{}',
            '^': r'\^{}',
            '\\': r'\textbackslash{}',
            '<': r'\textless ',
            '>': r'\textgreater ',
        }
        regex = re.compile(
            '|'.join(re.escape(key) for key in sorted(conversions.keys(), key=lambda item: - len(item))))
        return regex.sub(lambda match: conversions[match.group()], text)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys

from invoicing.command_line import Invoicing


def main():
    try:
        Invoicing()
    except KeyboardInterrupt:
        sys.exit()


if __name__ == "__main__":
    main()



from invoicing.query_builder.query_builder import QueryBuilder
from invoicing.repository.base_repository import BaseRepository


class ProjectRepository(BaseRepository):

    def __init__(self):
        super().__init__('project')

    def find_all_join_company(self):
        query = QueryBuilder(self.table) \
            .select(['project.id', 'reference_code', 'project.title'
                     'client.fullname as client_fullname',
                     'company.name as company_name',
                     'company.id as company_id',
                     'created_at']) \
            .from_() \
            .join('company', 'company_id = company.id')
        self.execute(**query.build())
        return self.get_all()

    def find_paginated(self, limit=5, page=1, **kwargs):
        query = QueryBuilder(self.table) \
            .select(['project.id', 'project.reference_code', 'project.title',
                     'company.name as company_name',
                     'created_at',
                     ]) \
            .from_() \
            .join('company', 'company_id = company.id') \
            .limit(limit) \
            .offset(page * limit - limit)
        self.execute(**query.build())
        return self.get_all()

    def find_paginated_join_company(self, limit=5, page=1):
        query = QueryBuilder(self.table) \
            .select(['project.id', 'reference_code', 'project.title',
                     'company.name as company_name',
                     'created_at',
                     ]) \
            .from_() \
            .join('company', 'company_id = company.id') \
            .limit(limit) \
            .offset(page * limit - limit)
        self.execute(**query.build())
        return self.get_all()

    def find_by_id_join_company(self, id):
        query = QueryBuilder(self.table) \
            .select(['project.id', 'reference_code', 'project.title',
                     'company.name as company_name',
                     'company.address as company_address',
                     'created_at',
                     ]) \
            .from_() \
            .join('company', 'company_id = company.id') \
            .where('project.id = ?', id)
        self.execute(**query.build())
        return self.get_one()

    def find_by_id_with_jobs(self, id):
        query = QueryBuilder(self.table) \
            .select(['project.id', 'projects.reference_code', 'project.title',
                     'company.name as company_name',
                     'company.address as company_address',
                     'company.id as company_id',
                     'job.title as job_title',
                     'job.description as job_description',
                     'staff.rate as rate', 'created_at']) \
            .from_() \
            .join('company', 'company_id = company.id') \
            .join('job', 'project.id = job.project_id') \
            .join('staff', 'job.assigned_to = staff.id') \
            .where('project.id = ?', id)
        self.execute(**query.build())
        return self.get_all()

    def find_projects_by_company_id(self, company_id):
        query = QueryBuilder(self.table) \
            .select(['id', 'reference_code', 'title', 'created_at']) \
            .from_() \
            .where('company_id = ?', company_id)
        self.execute(**query.build())
        return self.get_all()

    def find_last_inserted_id(self):
        query = QueryBuilder(self.table) \
            .select(['id']) \
            .from_() \
            .where('id = (select max(id) from projects)')
        self.execute(**query.build())
        return self.get_one()

    def find_last_reference_code(self):
        query = QueryBuilder(self.table) \
            .select(['reference_code as last_reference_code']) \
            .from_() \
            .where('id = (select max(id) from projects)')
        self.execute(**query.build())
        return self.get_one()

    def remove_projects_by_client_id(self, client_id):
        query = QueryBuilder(self.table) \
            .delete() \
            .where('client_id = ?', client_id)
        self.execute(**query.build())



from invoicing.crud.base_crud import BaseCrud
from invoicing.latex.latex_quote import LatexQuote
from invoicing.models.project_model import ProjectModel
from invoicing.repository.job_repository import JobRepository
from invoicing.repository.project_repository import ProjectRepository
from invoicing.ui.date import Date
from invoicing.ui.menu import Menu
from invoicing.ui.style import Style


class ProjectCrud(BaseCrud):
    def __init__(self):
        super().__init__('Projects', ProjectRepository, ProjectModel)
        self.menu_actions.add_action('Generate', self.generate)

    def make_paginated_menu(self):
        return self.paginated_menu(
            find=self.repository.find_paginated_join_company,
            find_by_id=self.repository.find_by_id_join_company
        )

    def generate(self):
        print(Style.create_title('Generate Quote'))
        project = self.paginated_menu(
            find=self.repository.find_paginated_join_company,
            find_by_id=self.repository.find_by_id_join_company
        )
        if project:
            jobs = JobRepository().find_jobs_by_project_id(project['id'])
            project_data = {
                'reference_code': project['reference_code'],
                'company_name': project['company_name'],
                'company_address': project['company_address'],
                'created_at': Date().convert_date_time_for_printing(project['created_at']),
                'total_cost': '£' + str(sum([float(job['rate']) * float(job['estimated_time']) for job in jobs])),
                'jobs': [{
                    'title': job['title'],
                    'description': job['description'],
                    'type': 'hours',
                    'estimated_time': str(job['estimated_time']),
                    'staff_rate': '£' + str(job['rate']),
                    'cost': '£' + str(float(job['rate']) * float(job['estimated_time']))
                } for job in jobs]
            }
            LatexQuote().generate(**project_data)
            Menu.wait_for_input()




from invoicing.api.base_api import BaseAPI
from invoicing.models.status_model import StatusModel
from invoicing.repository.status_repository import StatusRepository


class StatusAPI(BaseAPI):

    def __init__(self):
        super().__init__(StatusRepository(), StatusModel())



from invoicing.crud.base_crud import BaseCrud
from invoicing.models.client_model import ClientModel
from invoicing.repository.client_repository import ClientRepository


class ClientCrud(BaseCrud):
    def __init__(self):
        super().__init__('Clients', ClientRepository, ClientModel)

    def make_paginated_menu(self):
        return self.paginated_menu(
            find=self.repository.find_paginated_join_companies,
            find_by_id=self.repository.find_by_id_join_company
        )



from invoicing.api.base_api import BaseAPI
from invoicing.models.staff_model import StaffModel
from invoicing.repository.staff_repository import StaffRepository


class StaffAPI(BaseAPI):

    def __init__(self):
        super().__init__(StaffRepository(), StaffModel())



from invoicing.latex.latex_templating import LatexTemplating


class LatexInvoice(LatexTemplating):
    def generate(self, reference_code, company_name, company_address, created_at, total_cost, jobs):
        template = self.latex_jinja_env.get_template('templates/Invoice.example.tex')
        if len(jobs) == 0:
            jobs.append({
                'title': '-',
                'description': '-',
                'type': '-',
                'actual_time': '-',
                'staff_rate': '-',
                'cost': '-'
            })
        tex = template.render(
            reference_code=self.tex_escape(reference_code),
            company_name=self.tex_escape(company_name),
            company_address=self.tex_escape(company_address),
            date=self.tex_escape(created_at),
            total_cost=self.tex_escape(total_cost),
            jobs=[(lambda job: {k: self.tex_escape(v) for k, v in job.items()})(job) for job in jobs]
        )
        file_name = reference_code + " " + company_name
        self.create_tex_file(tex, file_name)
        self.create_pdf(file_name)


if __name__ == '__main__':
    LatexInvoice().generate(
        'I-7001',
        'Widget Corp',
        '100 Some street, A city, A town, BO41 0PN',
        '14/10/2018',
        '$160',
        [{
            'title': 'Job 1',
            'description': 'Do something',
            'type': 'service',
            'actual_time': '4h',
            'staff_rate': '$40',
            'cost': '$160'
        }]
    )



from invoicing.query_builder.query_builder import QueryBuilder
from invoicing.repository.base_repository import BaseRepository


class ClientRepository(BaseRepository):

    def __init__(self):
        super().__init__('client')

    def find_all_join_companies(self):
        query = QueryBuilder(self.table) \
            .select(['client.id', 'fullname', 'email', 'telephone', 'company.name as company_name']) \
            .from_() \
            .join('company', 'company_id = company.id')
        self.execute(**query.build())
        return self.get_all()

    def find_paginated_join_companies(self, limit=5, page=1):
        query = QueryBuilder(self.table) \
            .select(['client.id', 'fullname', 'email', 'telephone', 'company.name as company_name']) \
            .from_() \
            .join('company', 'company_id = company.id') \
            .limit(limit) \
            .offset(page * limit - limit)
        self.execute(**query.build())
        return self.get_all()

    def find_by_id_join_company(self, id):
        query = QueryBuilder(self.table) \
            .select(['client.id', 'fullname', 'email', 'telephone',
                     'company.name as company_name',
                     'company.address as company_address']) \
            .from_() \
            .join('company', 'company_id = company.id') \
            .where('client.id = ?', id)
        self.execute(**query.build())
        return self.get_one()

    def find_clients_by_company_id(self, company_id):
        query = QueryBuilder(self.table) \
            .select(['id', 'fullname', 'email', 'telephone', 'company_id']) \
            .from_() \
            .where('company_id = ?', company_id)
        self.execute(**query.build())
        return self.get_all()

    def find_client_and_company_by_id(self, id):
        query = QueryBuilder(self.table) \
            .select(['id', 'fullname', 'email', 'telephone', 'name', 'address']) \
            .from_() \
            .join('company', 'client.company_id = company.id') \
            .where('client.id = ?', id)
        self.execute(**query.build())
        return self.get_one()

    def remove_clients_by_company_id(self, company_id):
        query = QueryBuilder(self.table) \
            .delete() \
            .where('company_id = ?', company_id)
        self.execute(**query.build())



class Validation:
    def __init__(self, error_message='is not valid'):
        self.error_message = error_message

    def __call__(self, field):
        pass

    def validation_check(self, field, check):
        if not check:
            field.set_valid(False)
            field.set_error_message(self.error_message)
        else:
            field.set_valid(True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['invoicing',
 'invoicing.actions',
 'invoicing.api',
 'invoicing.constants',
 'invoicing.controller',
 'invoicing.crud',
 'invoicing.database',
 'invoicing.latex',
 'invoicing.model_validation',
 'invoicing.models',
 'invoicing.query_builder',
 'invoicing.relationships',
 'invoicing.repository',
 'invoicing.transformer',
 'invoicing.ui',
 'invoicing.value_validation']

package_data = \
{'': ['*'],
 'invoicing': ['sqlite/invoicing_bk.db',
               'sqlite/invoicing_bk.db',
               'sqlite/invoicing_dev.db',
               'sqlite/invoicing_dev.db',
               'templates/*']}

install_requires = \
['Flask==1.0.2',
 'Jinja2==2.10.3',
 'Pygments==2.2.0',
 'Werkzeug==0.16.0',
 'ansi-colours==0.2.6',
 'more-itertools==4.3.0',
 'pytest==3.10.0',
 'python-dateutil==2.7.5',
 'python-dotenv==0.9.1',
 'text-template==0.1.4',
 'uWSGI==2.0.17.1',
 'unicode==2.6']

entry_points = \
{'console_scripts': ['invoicing = invoicing.__main__:main']}

setup_kwargs = {
    'name': 'invoicing',
    'version': '2.1.0',
    'description': 'Generate invoice PDF from LaTeX template',
    'long_description': None,
    'author': 'Sarcoma',
    'author_email': 'sean@orderandchaoscreative.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



from invoicing.model_validation.field import StringField, ForeignKeyField
from invoicing.model_validation.validations import IsRequired
from invoicing.models.base_model import BaseModel
from invoicing.models.company_model import CompanyModel
from invoicing.relationships.base_relationship import BaseRelationship
from invoicing.repository.company_repository import CompanyRepository


class ClientModel(BaseModel):
    fullname = StringField([IsRequired()])
    email = StringField()
    telephone = StringField()
    company_id = ForeignKeyField(
        BaseRelationship('Company', CompanyRepository, CompanyModel)
    )



from ansi_colours import AnsiColours as Colour

from invoicing.repository.base_repository import BaseRepository
from invoicing.ui.menu import Menu
from invoicing.ui.style import Style
from invoicing.ui.table import Table
from invoicing.value_validation.value_validation import Validation


class Pagination:

    def __init__(self, repository: BaseRepository):
        self.repository = repository

    def __call__(self, limit=5, find=None, find_by_id=None):
        self.limit = limit
        self.find_paginated = self.repository.find_paginated if not find else find
        self.find_by_id = self.repository.find_by_id if not find_by_id else find_by_id
        self.total = self.repository.get_count()[0]
        self.page = 1
        return self.menu()

    def menu(self):
        result = False
        while not result:
            self.paginated_table()
            self.pagination_text()
            user_input = input('\nEnter an %s, use %s and %s to navigate, or %s to go back: ' % (
                Colour.green("'id'"),
                Colour.green("<"),
                Colour.green(">"),
                Colour.green("b")
            ))
            if (user_input == 'b'):
                break
            if self.should_change_page(user_input):
                self.change_page(user_input)
                continue
            if not Validation.is_number(user_input):
                print('Command not recognised')
                continue
            result = self.find_by_id(user_input)
            if not result:
                Menu.wait_for_input('An item with that ID was not found. Press any key to continue.')
        return result

    def paginated_table(self):
        Table.create_table(
            self.find_paginated(limit=self.limit, page=self.page),
            self.repository.get_headers()
        )

    def pagination_text(self):
        text = ""
        if self.page > 1:
            text += "%s previous | " % Colour.green('<')
        text += "page %d" % self.page
        if self.page * self.limit < self.total:
            text += " | next %s" % Colour.green('>')
        print(Style.create_underline(text))
        print(text)
        print(Style.create_underline(text))

    def should_change_page(self, user_input):
        return user_input is '<' or user_input is '>'

    def change_page(self, user_input):
        if self.page is not 1 and user_input is '<':
            self.page -= 1
        elif self.page * self.limit < self.total and user_input is '>':
            self.page += 1



from datetime import datetime

from invoicing.model_validation.validation import Validation


class IsRequired(Validation):
    def __init__(self, error_message='is required'):
        super().__init__(error_message)

    def __call__(self, field):
        self.validation_check(field, bool(field.value))
        return field


class IsString(Validation):
    def __init__(self, error_message='is not a string'):
        super().__init__(error_message)

    def __call__(self, field):
        self.validation_check(field, isinstance(field.value, str))
        return field


class IsInteger(Validation):
    def __init__(self, error_message='is not an integer'):
        super().__init__(error_message)

    def __call__(self, field):
        self.validation_check(field, isinstance(field.value, int))
        return field


class IsList(Validation):
    def __init__(self, error_message='is not a list'):
        super().__init__(error_message)

    def __call__(self, field):
        self.validation_check(field, isinstance(field.value, list))
        return field


class IsCallable(Validation):
    def __init__(self, error_message='is not a list'):
        super().__init__(error_message)

    def __call__(self, field):
        self.validation_check(field, True)
        return field

class IsBoolean(Validation):
    def __init__(self, error_message='is not a boolean'):
        super().__init__(error_message)

    def __call__(self, field):
        self.validation_check(field, isinstance(field.value, bool))
        return field


class IsFloat(Validation):
    def __init__(self, error_message='is not a float'):
        super().__init__(error_message)

    def __call__(self, field):
        self.validation_check(field, isinstance(field.value, float))
        return field


class MaxLength(Validation):
    def __init__(self, max, error_message='is too long'):
        super().__init__(error_message)
        self.max = max

    def __call__(self, field):
        self.validation_check(field, len(field.value) <= self.max)
        return field


class MinLength(Validation):
    def __init__(self, min, error_message='is too short'):
        super().__init__(error_message)
        self.min = min

    def __call__(self, field):
        self.validation_check(field, len(field.value) >= self.min)
        return field


class IsDate(Validation):
    def __init__(self, format="%d-%m-%y", error_message='is not a valid date'):
        super().__init__(error_message)
        self.format = format

    def __call__(self, field):
        self.validation_check(field, self.check_date(field.value))
        return field

    def check_date(self, value):
        try:
            datetime.strptime(value, self.format)
            return True
        except ValueError:
            return False



from dotenv import load_dotenv
import os
import invoicing.constants.constants as c

dotenv_path = os.path.join(os.path.dirname(__file__), '.env')
load_dotenv(dotenv_path)

DB = os.path.join(c.DB_PATH,os.environ.get("DB_FILE"))



from datetime import datetime

from invoicing.model_validation.field import Field, ForeignKeyField, OneToManyField
from invoicing.model_validation.validations import IsString
from invoicing.models.base_model import BaseModel
from invoicing.models.client_model import ClientModel
from invoicing.models.job_model import JobModel
from invoicing.relationships.base_relationship import BaseRelationship, OneToManyRelationship
from invoicing.repository.client_repository import ClientRepository
from invoicing.repository.job_repository import JobRepository
from invoicing.ui.pagination import Pagination


def job_paginated_menu():
    job_repository = JobRepository()
    return Pagination(job_repository)(
        find=job_repository.find_paginated_join_staff_and_status,
        find_by_id=lambda id: job_repository.find_by_id(id, (
            'id', 'reference_code', 'title',
            'estimated_time', 'actual_time', 'deadline',
            'assigned_to', 'status_id'
        )))


class InvoiceModel(BaseModel):
    client_id = ForeignKeyField(
        BaseRelationship('Client', ClientRepository, ClientModel)
    )
    job = OneToManyField(
        OneToManyRelationship('invoice_id', 'Job', JobRepository, JobModel, paginated_menu=job_paginated_menu)
    )
    created_at = Field([IsString()], initial_value=datetime.now().strftime("%Y-%m-%d %H:%M"), updatable=False)



from invoicing.crud.client_crud import ClientCrud
from invoicing.crud.company_crud import CompanyCrud
from invoicing.crud.invoice_crud import InvoiceCrud
from invoicing.crud.job_crud import JobCrud
from invoicing.crud.project_crud import ProjectCrud
from invoicing.crud.staff_crud import StaffCrud
from invoicing.crud.status_crud import StatusCrud


class MainController:

    @staticmethod
    def company_action():
        CompanyCrud().menu()

    @staticmethod
    def client_action():
        ClientCrud().menu()

    @staticmethod
    def project_action():
        ProjectCrud().menu()

    @staticmethod
    def invoice_action():
        InvoiceCrud().menu()

    @staticmethod
    def job_action():
        JobCrud().menu()

    @staticmethod
    def staff_action():
        StaffCrud().menu()

    @staticmethod
    def status_action():
        StatusCrud().menu()



from invoicing.actions.action import Action


class ActionCollection:

    def __init__(self, *args, exit_action=None):
        self.actions = []
        self.count = 0
        for arg in args:
            self.count += 1
            self.actions.append(Action(str(self.count), arg[0], arg[1]))
        self.actions.append(exit_action or Action('b', 'Back', False))

    def add_action(self, label, action):
        self.count += 1
        self.actions.insert(-1, Action(str(self.count), label, action))

    def add_actions(self, *args):
        for arg in reversed(args):
            self.count += 1
            self.actions.insert(-1, Action(str(self.count), arg[0], arg[1]))



import collections

from invoicing.model_validation.field import Field


class OrderedClassMembers(type):
    @classmethod
    def __prepare__(self, name, bases):
        return collections.OrderedDict()

    def __new__(cls, name, bases, class_dict):
        result = type.__new__(cls, name, bases, dict(class_dict))
        exclude = set(dir(type))
        result.__fields__ = list(f for f in class_dict.keys() if f not in exclude)
        return result


class BaseModel(metaclass=OrderedClassMembers):
    valid = False
    errors = {}
    fields = None

    def __call__(self, **kwargs):
        self.valid = False
        self.errors = {}
        for field, value in kwargs.items():
            attributes = self.__class__.__dict__
            if field in attributes and isinstance(attributes[field], Field):
                attributes[field].set_value(value)

    def __iter__(self):
        for field in self.__fields__:
            if isinstance(self.__class__.__dict__[field], Field):
                yield (field, self.__class__.__dict__[field])

    def validate(self):
        for field, value in self:
            value.validate()
            if not value.is_valid():
                self.errors[field] = value.get_error_message()
        self.valid = False if len(self.errors) else True

    def is_valid(self):
        return self.valid

    def get_errors(self):
        return self.errors

    def __str__(self):
        return "Errors: %s" % ', '.join([k + ": " + v for k, v in self.errors.items()]) if len(self.errors) else 'Valid'
